/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 21/04/22.
//

#ifndef FGGL_GUI_FONTS_HPP
#define FGGL_GUI_FONTS_HPP

#include <memory>
#include <map>
#include <string>

#include "fggl/math/types.hpp"
#include "fggl/data/storage.hpp"
#include "fggl/modules/module.hpp"

#include <ft2build.h>
#include FT_FREETYPE_H


namespace fggl::gui {
	constexpr const char* DEFAULT_FONT_NAME = "LiberationSans-Regular.ttf";


	struct GlyphMetrics {
		math::vec2 size;
		math::vec2 bearing;
		long advance;
	};

	class FontFace {
		public:
			explicit FontFace(FT_Face face);
			~FontFace();

			FontFace(const FontFace &) = delete;
			FontFace(FontFace &&) = delete;
			FontFace &operator=(const FontFace &) = delete;
			FontFace &operator=(FontFace &&) = delete;

			inline GlyphMetrics &metrics(char letter) {
				auto itr = m_metrics.find(letter);
				if (itr == m_metrics.end()) {
					return populateMetrics(letter);
				}
				return itr->second;
			}

			math::vec2 stringSize(const std::string &text);
			void texture(char letter, int &width, int &height, unsigned char **buff);

		private:
			FT_Face m_face;
			std::map<char, GlyphMetrics> m_metrics;

			GlyphMetrics &populateMetrics(char letter);
	};

	class FontLibrary {
		public:
			constexpr static const modules::ModuleService service = modules::make_service("fggl::gui::font");

			explicit FontLibrary(data::Storage *storage);
			~FontLibrary();

			// copy and moving not needed
			FontLibrary(const FontLibrary &) = delete;
			FontLibrary(FontLibrary &&) = delete;
			FontLibrary &operator=(const FontLibrary &) = delete;
			FontLibrary &operator=(FontLibrary &&) = delete;

			inline std::shared_ptr<FontFace> getFont(const std::string &name) {
				auto fontItr = m_cache.find(name);
				if (fontItr != m_cache.end()) {
					return fontItr->second;
				}

				// need to load the font...
				auto path = m_storage->resolvePath(data::StorageType::Data, name);

				FT_Face face;
				if (FT_New_Face(m_context, path.string().c_str(), 0, &face)) {
					return nullptr;
				}
				FT_Set_Pixel_Sizes(face, 0, 18);

				// create the new font object
				auto ptr = std::make_shared<FontFace>(face);
				m_cache[name] = ptr;
				return ptr;
			}

			inline void setDefaultFont(const std::string& name) {
				m_defaultFont = getFont(name);
			}

			inline std::shared_ptr<FontFace> getDefaultFont() const {
				return m_defaultFont;
			}
		private:
			FT_Library m_context;
			data::Storage *m_storage;
			std::map<const std::string, std::shared_ptr<FontFace>> m_cache;
			std::shared_ptr<FontFace> m_defaultFont;
	};

} // nmespace fggl::gui

#endif //FGGL_GUI_FONTS_HPP
